

; (in-package "IPASIR")

(include-book "centaur/fty/deftypes" :dir :system)
(include-book "std/util/defenum" :dir :system)
(include-book "std/stobjs/absstobjs" :dir :system)

(local (std::add-default-post-define-hook :fix))

(define litp (x)
  (and (integerp x) (not (equal x 0)))
  ///
  (defthm litp-compound-recognizer
    (equal (litp x)
           (and (integerp x) (not (equal x 0))))
    :rule-classes :compound-recognizer)

  (define lit-fix ((x litp))
    :returns (new-x litp :rule-classes :type-prescription)
    (mbe :logic (if (litp x) x 1)
         :exec x)
    ///
    (defthm lit-fix-when-litp
      (implies (litp x)
               (equal (lit-fix x) x)))

    (fty::deffixtype lit :pred litp :fix lit-fix :equiv lit-equiv
      :define t :forward t)))

;; also represents an assignment -- really just a lit list
(fty::deflist clause :pred clausep :elt-type lit :true-listp t)

(fty::deflist formula :pred formulap :elt-type clause :true-listp t)

(std::defenum ipasir-status-p
  (:undef :input :unsat :sat))

;; Semantics ignoring don't-cares
(define lit-eval ((lit litp)
                  (assignment clausep))
  :measure (len assignment)
  (if (atom assignment)
      ;; positive lits are false, negative are true, if not present
      (< (lit-fix lit) 0)
    (b* ((lit (lit-fix lit))
         (lit1 (lit-fix (car assignment))))
      (cond ((equal lit lit1) t)
            ((equal lit (- lit1)) nil)
            (t (lit-eval lit (cdr assignment))))))
  ///
  (defthm lit-eval-of-negate
    (implies (litp lit)
             (equal (lit-eval (- lit) assignment)
                    (not (lit-eval lit assignment))))))

(define clause-eval ((clause clausep)
                     (assignment clausep))
  :measure (len clause)
  (if (atom clause)
      nil
    (or (lit-eval (car clause) assignment)
        (clause-eval (cdr clause) assignment))))

(define assumption-eval ((assumption clausep)
                     (assignment clausep))
  :measure (len assumption)
  (if (atom assumption)
      t
    (and (lit-eval (car assumption) assignment)
         (assumption-eval (cdr assumption) assignment))))

(define formula-eval ((formula formulap)
                     (assignment clausep))
  :measure (len formula)
  (if (atom formula)
      t
    (and (clause-eval (car formula) assignment)
         (formula-eval (cdr formula) assignment))))
      


(fty::defprod ipasir-solver$a
  ((formula formulap    "Permanent formula")
   (assumption clausep  "Current assumption, if status is :input, or assumption
                         before latest solve, if status is :unsat.")
   (new-clause clausep  "Partial clause to add to the formula")
   (solution   clausep  "Satisfying assignment from solver, when status = :sat,
                         or subset of assumptions sufficient to prove unsat, when
                         status = :unsat")
   (status ipasir-status-p :default ':undef)
   (signature stringp :rule-classes :type-prescription
              "Name of the solver library")))

(define ipasir-signature$a ((solver ipasir-solver$a-p))
  :enabled t
  (ipasir-solver$a->signature solver))

(define ipasir-init$a ((solver ipasir-solver$a-p))
  :guard (eq (ipasir-solver$a->status solver) :undef)
  :returns (new-solver ipasir-solver$a-p)
  (change-ipasir-solver$a solver :status :input)
  ///
  (std::defret ipasir-init$a-status
    (equal (ipasir-solver$a->status new-solver) :input)))

(define ipasir-release$a ((solver ipasir-solver$a-p))
  :guard (not (eq (ipasir-solver$a->status solver) :undef))
  :returns (new-solver ipasir-solver$a-p)
  (change-ipasir-solver$a solver :status :undef)
  ///
  (std::defret ipasir-release$a-status
    (equal (ipasir-solver$a->status new-solver) :undef)))

(define ipasir-add$a ((solver ipasir-solver$a-p)
                      (lit-or-zero integerp))
  :guard (not (eq (ipasir-solver$a->status solver) :undef))
  :returns (new-solver ipasir-solver$a-p)
  (b* (((ipasir-solver$a solver))
       (lit-or-zero (acl2::lifix lit-or-zero))
       ((when (eql lit-or-zero 0))
        ;; finalize the current new-clause by adding it to the formula and emptying it
        (change-ipasir-solver$a solver :formula (cons solver.new-clause solver.formula)
                                :new-clause nil
                                :solution nil
                                :status :input))
       (- (mbe :logic nil
               :exec (and (not (signed-byte-p 32 lit-or-zero))
                          (raise "Out of bounds literal: ~x0" lit-or-zero)))))
    ;; add literal to first clause in formula
    (change-ipasir-solver$a solver
                            :new-clause (cons lit-or-zero solver.new-clause)
                            :solution nil
                            :status :input))
  ///
  (std::defret status-of-ipasir-add$a
    (equal (ipasir-solver$a->status new-solver) :input)))

(define ipasir-assume$a ((solver ipasir-solver$a-p)
                         (lit litp))
  :guard (not (eq (ipasir-solver$a->status solver) :undef))
  :returns (new-solver ipasir-solver$a-p)
  (b* (((ipasir-solver$a solver))
       ;; if status is unsat, we've kept around the previous assumption until
       ;; now for querying with ipasir-failed, but now we need to empty it.
       (assumption (if (eq solver.status :unsat) nil solver.assumption)))
    (change-ipasir-solver$a solver :assumption (cons (lit-fix lit) assumption)
                            :solution nil
                            :status :input))
  ///
  (std::defret status-of-ipasir-assume$a
    (equal (ipasir-solver$a->status new-solver) :input)))

(define ipasir-val$a ((solver ipasir-solver$a-p)
                      (lit litp))
  :guard (eq (ipasir-solver$a->status solver) :sat)
  :Returns (lit-or-zero integerp)
  (b* (((ipasir-solver$a solver))
       (lit (lit-fix lit)))
    (cond ((member lit solver.solution) lit)
          ((member (- lit) solver.solution) (- lit))
          (t 0))))

(define ipasir-failed$a ((solver ipasir-solver$a-p)
                         (lit litp))
  :guard (b* (((ipasir-solver$a solver)))
           (and (eq solver.status :unsat)
                (member lit solver.assumption)))
  (b* (((ipasir-solver$a solver))
       (lit (lit-fix lit)))
    (if (member lit solver.solution) 1 0)))


(defun ipasir-useless-clauseproc (clause)
  (list clause))

(defttag ipasir-solve$a)

(define-trusted-clause-processor
  ipasir-useless-clauseproc
  (ipasir-solve$a)
  :partial-theory
  (encapsulate
    (((ipasir-solve$a *) => (mv * *)
      :guard (and (ipasir-solver$a-p solver)
                  (not (equal (ipasir-solver$a->status solver) :undef)))
      :formals (solver)))

    (local (define ipasir-solve$a ((solver ipasir-solver$a-p))
             :enabled t
             ;; returns (mv status solver)
             (mv :failed
                 (change-ipasir-solver$a solver
                                         :solution nil
                                         :assumption nil
                                         :status :input))))

    (defthm ipasir-solve$a-general
      (b* (((ipasir-solver$a solver))
           ((mv ?status (ipasir-solver$a new-solver)) (ipasir-solve$a solver)))
        (and (equal new-solver.formula solver.formula)
             (equal new-solver.signature solver.signature)
             (not (equal new-solver.status :undef))
             (ipasir-solver$a-p new-solver))))


    (defthm ipasir-solve$a-unsat
      (b* (((ipasir-solver$a solver))
           ((mv status (ipasir-solver$a new-solver)) (ipasir-solve$a solver)))
        (implies (equal status :unsat)
                 (and (subsetp new-solver.solution solver.assumption)
                      (equal new-solver.assumption solver.assumption)
                      (equal new-solver.status :unsat)
                      ;; no environment exists which satisfies the assumption (as an assignment) and the formula
                      (implies (formula-eval solver.formula assignment)
                               (not (assumption-eval new-solver.solution assignment)))))))

    (defthm ipasir-solve$a-sat
      (b* (((ipasir-solver$a solver))
           ((mv status (ipasir-solver$a new-solver)) (ipasir-solve$a solver)))
        (implies (equal status :sat)
                 (and (equal new-solver.status :sat)
                      (equal new-solver.assumption nil)))))
    ;; Won't assume the solution is a correct assignment because we can just check it.

    (defthm ipasir-solve$a-failed
      (b* (((ipasir-solver$a solver))
           ((mv status (ipasir-solver$a new-solver)) (ipasir-solve$a solver)))
        (implies (equal status :failed)
                 (and (equal new-solver.status :input)
                      (equal new-solver.assumption nil)))))))

(local (in-theory (disable ifix nth update-nth)))

(make-event
 `(defstobj ipasir-fake-solver$c
    (ipasir-fake-solver-val
     :type (satisfies ipasir-solver$a-p)
     :initially ,(make-ipasir-solver$a :status :undef :signature "Fake"))
    :renaming ((ipasir-fake-solver-val        ipasir-fake-solver-get)
               (update-ipasir-fake-solver-val ipasir-fake-solver-set))))

(define create-ipasir-fake-solver$a ()
  :enabled t
  (make-ipasir-solver$a :status :undef :signature "Fake"))

(define ipasir-fake-signature$c (ipasir-fake-solver$c)
  :enabled t
  (ipasir-signature$a (ipasir-fake-solver-get ipasir-fake-solver$c)))

(define ipasir-fake-init$c (ipasir-fake-solver$c)
  :guard (eq (ipasir-solver$a->status (ipasir-fake-solver-get ipasir-fake-solver$c)) :undef)
  :enabled t
  (ipasir-fake-solver-set (ipasir-init$a (ipasir-fake-solver-get ipasir-fake-solver$c)) ipasir-fake-solver$c))

(define ipasir-fake-release$c (ipasir-fake-solver$c)
  :guard (not (eq (ipasir-solver$a->status (ipasir-fake-solver-get ipasir-fake-solver$c)) :undef))
  :enabled t
  (ipasir-fake-solver-set (ipasir-release$a (ipasir-fake-solver-get ipasir-fake-solver$c)) ipasir-fake-solver$c))


(define ipasir-fake-add$c (ipasir-fake-solver$c (lit-or-zero integerp))
  :guard (not (eq (ipasir-solver$a->status (ipasir-fake-solver-get ipasir-fake-solver$c)) :undef))
  :enabled t
  (ipasir-fake-solver-set (ipasir-add$a (ipasir-fake-solver-get ipasir-fake-solver$c) lit-or-zero) ipasir-fake-solver$c))

(define ipasir-fake-assume$c (ipasir-fake-solver$c (lit litp))
  :guard (not (eq (ipasir-solver$a->status (ipasir-fake-solver-get ipasir-fake-solver$c)) :undef))
  :enabled t
  (ipasir-fake-solver-set (ipasir-assume$a (ipasir-fake-solver-get ipasir-fake-solver$c) lit) ipasir-fake-solver$c))

(define ipasir-fake-val$c (ipasir-fake-solver$c (lit litp))
  :guard (eq (ipasir-solver$a->status (ipasir-fake-solver-get ipasir-fake-solver$c)) :sat)
  :enabled t
  (ipasir-val$a (ipasir-fake-solver-get ipasir-fake-solver$c) lit))

(define ipasir-fake-failed$c (ipasir-fake-solver$c (lit litp))
  :guard (and (eq (ipasir-solver$a->status (ipasir-fake-solver-get ipasir-fake-solver$c)) :unsat)
              (member lit (ipasir-solver$a->assumption (ipasir-fake-solver-get ipasir-fake-solver$c))))
  :enabled t
  (ipasir-failed$a (ipasir-fake-solver-get ipasir-fake-solver$c) lit))

(define ipasir-fake-solve$c (ipasir-fake-solver$c)
  :guard (not (eq (ipasir-solver$a->status (ipasir-fake-solver-get ipasir-fake-solver$c)) :undef))
  :enabled t
  (b* (((mv status solver) (ipasir-solve$a (ipasir-fake-solver-get ipasir-fake-solver$c)))
       (ipasir-fake-solver$c (ipasir-fake-solver-set solver ipasir-fake-solver$c)))
    (mv status ipasir-fake-solver$c)))



(local (define ipasir-fake-corr (ipasir-fake-solver$c solver)
         :enabled t
         (equal (ipasir-fake-solver-get ipasir-fake-solver$c) solver)))

#||
(acl2::defabsstobj-events ipasir-fake-solver
  :concrete ipasir-fake-solver$c
  :recognizer (ipasir-fake-solverp :logic ipasir-solver$a-p :exec ipasir-fake-solver$cp)
  :creator (crete-ipasir-fake-solver :logic create-ipasir-fake-solver$a
                                     :exec create-ipasir-fake-solver$c)
  :corr-fn ipasir-fake-corr
  :exports ((ipasir-fake-signature :logic ipasir-signature$a :exec ipasir-fake-signature$c)
            (ipasir-fake-init :logic ipasir-init$a :exec ipasir-fake-init$c)
            (ipasir-fake-release :logic ipasir-release$a :exec ipasir-fake-release$c)
            (ipasir-fake-add :logic ipasir-add$a :exec ipasir-fake-add$c)
            (ipasir-fake-assume :logic ipasir-assume$a :exec ipasir-fake-assume$c)
            (ipasir-fake-val :logic ipasir-val$a :exec ipasir-fake-val$c)
            (ipasir-fake-failed :logic ipasir-failed$a :exec ipasir-fake-failed$c)
            (ipasir-fake-solve :logic ipasir-solve$a :exec ipasir-fake-solve$c)))

||#

(DEFTHM CRETE-IPASIR-FAKE-SOLVER{CORRESPONDENCE}
        (IPASIR-FAKE-CORR (CREATE-IPASIR-FAKE-SOLVER$C)
                          (CREATE-IPASIR-FAKE-SOLVER$A))
        :RULE-CLASSES NIL)
(DEFTHM CRETE-IPASIR-FAKE-SOLVER{PRESERVED}
        (IPASIR-SOLVER$A-P (CREATE-IPASIR-FAKE-SOLVER$A))
        :RULE-CLASSES NIL)
(DEFTHM
   IPASIR-FAKE-SIGNATURE{CORRESPONDENCE}
   (IMPLIES (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
                 (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER))
            (EQUAL (IPASIR-FAKE-SIGNATURE$C IPASIR-FAKE-SOLVER$C)
                   (IPASIR-SIGNATURE$A IPASIR-FAKE-SOLVER)))
   :RULE-CLASSES NIL)
(DEFTHM
   IPASIR-FAKE-INIT{CORRESPONDENCE}
   (IMPLIES (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
                 (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
                 (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                     :UNDEF))
            (IPASIR-FAKE-CORR (IPASIR-FAKE-INIT$C IPASIR-FAKE-SOLVER$C)
                              (IPASIR-INIT$A IPASIR-FAKE-SOLVER)))
   :RULE-CLASSES NIL)
(DEFTHM
 IPASIR-FAKE-INIT{GUARD-THM}
 (IMPLIES
  (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
       (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
       (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
           :UNDEF))
  (EQ
    (IPASIR-SOLVER$A->STATUS (IPASIR-FAKE-SOLVER-GET IPASIR-FAKE-SOLVER$C))
    :UNDEF))
 :RULE-CLASSES NIL)
(DEFTHM IPASIR-FAKE-INIT{PRESERVED}
        (IMPLIES (AND (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
                      (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                          :UNDEF))
                 (IPASIR-SOLVER$A-P (IPASIR-INIT$A IPASIR-FAKE-SOLVER)))
        :RULE-CLASSES NIL)
(DEFTHM
   IPASIR-FAKE-RELEASE{CORRESPONDENCE}
   (IMPLIES (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
                 (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
                 (NOT (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                          :UNDEF)))
            (IPASIR-FAKE-CORR (IPASIR-FAKE-RELEASE$C IPASIR-FAKE-SOLVER$C)
                              (IPASIR-RELEASE$A IPASIR-FAKE-SOLVER)))
   :RULE-CLASSES NIL)
(DEFTHM
 IPASIR-FAKE-RELEASE{GUARD-THM}
 (IMPLIES
  (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
       (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
       (NOT (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                :UNDEF)))
  (NOT
   (EQ
    (IPASIR-SOLVER$A->STATUS (IPASIR-FAKE-SOLVER-GET IPASIR-FAKE-SOLVER$C))
    :UNDEF)))
 :RULE-CLASSES NIL)
(DEFTHM IPASIR-FAKE-RELEASE{PRESERVED}
        (IMPLIES (AND (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
                      (NOT (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                               :UNDEF)))
                 (IPASIR-SOLVER$A-P (IPASIR-RELEASE$A IPASIR-FAKE-SOLVER)))
        :RULE-CLASSES NIL)
(DEFTHM
 IPASIR-FAKE-ADD{CORRESPONDENCE}
 (IMPLIES
     (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
          (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
          (INTEGERP LIT-OR-ZERO)
          (NOT (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                   :UNDEF)))
     (IPASIR-FAKE-CORR (IPASIR-FAKE-ADD$C IPASIR-FAKE-SOLVER$C LIT-OR-ZERO)
                       (IPASIR-ADD$A IPASIR-FAKE-SOLVER LIT-OR-ZERO)))
 :RULE-CLASSES NIL)
(DEFTHM
 IPASIR-FAKE-ADD{GUARD-THM}
 (IMPLIES (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
               (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
               (INTEGERP LIT-OR-ZERO)
               (NOT (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                        :UNDEF)))
          (AND (INTEGERP LIT-OR-ZERO)
               (NOT (EQ (IPASIR-SOLVER$A->STATUS
                             (IPASIR-FAKE-SOLVER-GET IPASIR-FAKE-SOLVER$C))
                        :UNDEF))))
 :RULE-CLASSES NIL)
(DEFTHM
    IPASIR-FAKE-ADD{PRESERVED}
    (IMPLIES
         (AND (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
              (INTEGERP LIT-OR-ZERO)
              (NOT (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                       :UNDEF)))
         (IPASIR-SOLVER$A-P (IPASIR-ADD$A IPASIR-FAKE-SOLVER LIT-OR-ZERO)))
    :RULE-CLASSES NIL)
(DEFTHM
 IPASIR-FAKE-ASSUME{CORRESPONDENCE}
 (IMPLIES (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
               (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
               (LITP LIT)
               (NOT (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                        :UNDEF)))
          (IPASIR-FAKE-CORR (IPASIR-FAKE-ASSUME$C IPASIR-FAKE-SOLVER$C LIT)
                            (IPASIR-ASSUME$A IPASIR-FAKE-SOLVER LIT)))
 :RULE-CLASSES NIL)
(DEFTHM
 IPASIR-FAKE-ASSUME{GUARD-THM}
 (IMPLIES (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
               (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
               (LITP LIT)
               (NOT (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                        :UNDEF)))
          (AND (LITP LIT)
               (NOT (EQ (IPASIR-SOLVER$A->STATUS
                             (IPASIR-FAKE-SOLVER-GET IPASIR-FAKE-SOLVER$C))
                        :UNDEF))))
 :RULE-CLASSES NIL)
(DEFTHM
     IPASIR-FAKE-ASSUME{PRESERVED}
     (IMPLIES (AND (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
                   (LITP LIT)
                   (NOT (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                            :UNDEF)))
              (IPASIR-SOLVER$A-P (IPASIR-ASSUME$A IPASIR-FAKE-SOLVER LIT)))
     :RULE-CLASSES NIL)
(DEFTHM
   IPASIR-FAKE-VAL{CORRESPONDENCE}
   (IMPLIES (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
                 (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
                 (LITP LIT)
                 (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                     :SAT))
            (EQUAL (IPASIR-FAKE-VAL$C IPASIR-FAKE-SOLVER$C LIT)
                   (IPASIR-VAL$A IPASIR-FAKE-SOLVER LIT)))
   :RULE-CLASSES NIL)
(DEFTHM
 IPASIR-FAKE-VAL{GUARD-THM}
 (IMPLIES
  (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
       (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
       (LITP LIT)
       (EQ (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
           :SAT))
  (AND
   (LITP LIT)
   (EQ
    (IPASIR-SOLVER$A->STATUS (IPASIR-FAKE-SOLVER-GET IPASIR-FAKE-SOLVER$C))
    :SAT)))
 :RULE-CLASSES NIL)
(DEFTHM
 IPASIR-FAKE-FAILED{CORRESPONDENCE}
 (IMPLIES
  (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
       (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
       (LITP LIT)
       (LET ((SOLVER.STATUS (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER))
             (SOLVER IPASIR-FAKE-SOLVER))
            (LET ((SOLVER.ASSUMPTION (IPASIR-SOLVER$A->ASSUMPTION SOLVER)))
                 (AND (EQ SOLVER.STATUS :UNSAT)
                      (LET ((X LIT) (L SOLVER.ASSUMPTION))
                           (MBE1 (MEMBER-EQL-EXEC X L)
                                 (PROG2$ (MEMBER-EQL-EXEC$GUARD-CHECK X L)
                                         (MEMBER-EQUAL X L))))))))
  (EQUAL (IPASIR-FAKE-FAILED$C IPASIR-FAKE-SOLVER$C LIT)
         (IPASIR-FAILED$A IPASIR-FAKE-SOLVER LIT)))
 :RULE-CLASSES NIL)
(DEFTHM
 IPASIR-FAKE-FAILED{GUARD-THM}
 (IMPLIES
  (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
       (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
       (LITP LIT)
       (LET ((SOLVER.STATUS (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER))
             (SOLVER IPASIR-FAKE-SOLVER))
            (LET ((SOLVER.ASSUMPTION (IPASIR-SOLVER$A->ASSUMPTION SOLVER)))
                 (AND (EQ SOLVER.STATUS :UNSAT)
                      (LET ((X LIT) (L SOLVER.ASSUMPTION))
                           (MBE1 (MEMBER-EQL-EXEC X L)
                                 (PROG2$ (MEMBER-EQL-EXEC$GUARD-CHECK X L)
                                         (MEMBER-EQUAL X L))))))))
  (AND
   (LITP LIT)
   (EQ
    (IPASIR-SOLVER$A->STATUS (IPASIR-FAKE-SOLVER-GET IPASIR-FAKE-SOLVER$C))
    :UNSAT)
   (LET ((X LIT)
         (L (IPASIR-SOLVER$A->ASSUMPTION
                 (IPASIR-FAKE-SOLVER-GET IPASIR-FAKE-SOLVER$C))))
        (MBE1 (MEMBER-EQL-EXEC X L)
              (PROG2$ (MEMBER-EQL-EXEC$GUARD-CHECK X L)
                      (MEMBER-EQUAL X L))))))
 :RULE-CLASSES NIL)
(DEFTHM
   IPASIR-FAKE-SOLVE{CORRESPONDENCE}
   (IMPLIES (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
                 (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
                 (NOT (EQUAL (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                             :UNDEF)))
            (LET ((LHS (IPASIR-FAKE-SOLVE$C IPASIR-FAKE-SOLVER$C))
                  (RHS (IPASIR-SOLVE$A IPASIR-FAKE-SOLVER)))
                 (AND (EQUAL (MV-NTH 0 LHS) (MV-NTH 0 RHS))
                      (IPASIR-FAKE-CORR (MV-NTH 1 LHS)
                                        (MV-NTH 1 RHS)))))
   :RULE-CLASSES NIL)
(DEFTHM
 IPASIR-FAKE-SOLVE{GUARD-THM}
 (IMPLIES
  (AND (IPASIR-FAKE-CORR IPASIR-FAKE-SOLVER$C IPASIR-FAKE-SOLVER)
       (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
       (NOT (EQUAL (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                   :UNDEF)))
  (NOT
   (EQ
    (IPASIR-SOLVER$A->STATUS (IPASIR-FAKE-SOLVER-GET IPASIR-FAKE-SOLVER$C))
    :UNDEF)))
 :RULE-CLASSES NIL)
(DEFTHM
   IPASIR-FAKE-SOLVE{PRESERVED}
   (IMPLIES
        (AND (IPASIR-SOLVER$A-P IPASIR-FAKE-SOLVER)
             (NOT (EQUAL (IPASIR-SOLVER$A->STATUS IPASIR-FAKE-SOLVER)
                         :UNDEF)))
        (IPASIR-SOLVER$A-P (MV-NTH 1 (IPASIR-SOLVE$A IPASIR-FAKE-SOLVER))))
   :RULE-CLASSES NIL)

(DEFABSSTOBJ
       IPASIR-FAKE-SOLVER
       :CONCRETE IPASIR-FAKE-SOLVER$C
       :RECOGNIZER (IPASIR-FAKE-SOLVERP :LOGIC IPASIR-SOLVER$A-P
                                        :EXEC IPASIR-FAKE-SOLVER$CP)
       :CREATOR (CRETE-IPASIR-FAKE-SOLVER :LOGIC CREATE-IPASIR-FAKE-SOLVER$A
                                          :EXEC CREATE-IPASIR-FAKE-SOLVER$C)
       :CORR-FN IPASIR-FAKE-CORR
       :EXPORTS ((IPASIR-FAKE-SIGNATURE :LOGIC IPASIR-SIGNATURE$A
                                        :EXEC IPASIR-FAKE-SIGNATURE$C)
                 (IPASIR-FAKE-INIT :LOGIC IPASIR-INIT$A
                                   :EXEC IPASIR-FAKE-INIT$C)
                 (IPASIR-FAKE-RELEASE :LOGIC IPASIR-RELEASE$A
                                      :EXEC IPASIR-FAKE-RELEASE$C)
                 (IPASIR-FAKE-ADD :LOGIC IPASIR-ADD$A
                                  :EXEC IPASIR-FAKE-ADD$C)
                 (IPASIR-FAKE-ASSUME :LOGIC IPASIR-ASSUME$A
                                     :EXEC IPASIR-FAKE-ASSUME$C)
                 (IPASIR-FAKE-VAL :LOGIC IPASIR-VAL$A
                                  :EXEC IPASIR-FAKE-VAL$C)
                 (IPASIR-FAKE-FAILED :LOGIC IPASIR-FAILED$A
                                     :EXEC IPASIR-FAKE-FAILED$C)
                 (IPASIR-FAKE-SOLVE :LOGIC IPASIR-SOLVE$A
                                    :EXEC IPASIR-FAKE-SOLVE$C)))
